from datetime import datetime
from urllib.parse import quote

from ..exceptions import AnymailError, AnymailRequestsAPIError
from ..message import AnymailRecipientStatus
from ..utils import get_anymail_setting, rfc2822date
from .base_requests import AnymailRequestsBackend, RequestsPayload


class EmailBackend(AnymailRequestsBackend):
    """
    Mailgun API Email Backend
    """

    esp_name = "Mailgun"

    def __init__(self, **kwargs):
        """Init options from Django settings"""
        esp_name = self.esp_name
        self.api_key = get_anymail_setting(
            "api_key", esp_name=esp_name, kwargs=kwargs, allow_bare=True
        )
        self.sender_domain = get_anymail_setting(
            "sender_domain",
            esp_name=esp_name,
            kwargs=kwargs,
            allow_bare=True,
            default=None,
        )
        api_url = get_anymail_setting(
            "api_url",
            esp_name=esp_name,
            kwargs=kwargs,
            default="https://api.mailgun.net/v3",
        )
        if not api_url.endswith("/"):
            api_url += "/"
        super().__init__(api_url, **kwargs)

    def build_message_payload(self, message, defaults):
        return MailgunPayload(message, defaults, self)

    def raise_for_status(self, response, payload, message):
        # Mailgun issues a terse 404 for unrecognized sender domains.
        # Add some context:
        if response.status_code == 404 and "Domain not found" in response.text:
            raise AnymailRequestsAPIError(
                "Unknown sender domain {sender_domain!r}.\n"
                "Check the domain is verified with Mailgun, and that the ANYMAIL"
                " MAILGUN_API_URL setting {api_url!r} is the correct region.".format(
                    sender_domain=payload.sender_domain, api_url=self.api_url
                ),
                email_message=message,
                payload=payload,
                response=response,
                backend=self,
            )

        super().raise_for_status(response, payload, message)

        # Mailgun issues a cryptic "Mailgun Magnificent API" success response
        # for invalid API endpoints. Convert that to a useful error:
        if response.status_code == 200 and "Mailgun Magnificent API" in response.text:
            raise AnymailRequestsAPIError(
                "Invalid Mailgun API endpoint %r.\n"
                "Check your ANYMAIL MAILGUN_SENDER_DOMAIN"
                " and MAILGUN_API_URL settings." % response.url,
                email_message=message,
                payload=payload,
                response=response,
                backend=self,
            )

    def parse_recipient_status(self, response, payload, message):
        # The *only* 200 response from Mailgun seems to be:
        #     {
        #       "id": "<20160306015544.116301.25145@example.org>",
        #       "message": "Queued. Thank you."
        #     }
        #
        # That single message id applies to all recipients.
        # The only way to detect rejected, etc. is via webhooks.
        # (*Any* invalid recipient addresses will generate a 400 API error)
        parsed_response = self.deserialize_json_response(response, payload, message)
        try:
            message_id = parsed_response["id"]
            mailgun_message = parsed_response["message"]
        except (KeyError, TypeError) as err:
            raise AnymailRequestsAPIError(
                "Invalid Mailgun API response format",
                email_message=message,
                payload=payload,
                response=response,
                backend=self,
            ) from err
        if not mailgun_message.startswith("Queued"):
            raise AnymailRequestsAPIError(
                "Unrecognized Mailgun API message '%s'" % mailgun_message,
                email_message=message,
                payload=payload,
                response=response,
                backend=self,
            )
        # Simulate a per-recipient status of "queued":
        status = AnymailRecipientStatus(message_id=message_id, status="queued")
        return {recipient.addr_spec: status for recipient in payload.all_recipients}


class MailgunPayload(RequestsPayload):
    def __init__(self, message, defaults, backend, *args, **kwargs):
        auth = ("api", backend.api_key)
        self.sender_domain = backend.sender_domain
        self.all_recipients = []  # used for backend.parse_recipient_status

        # late-binding of recipient-variables:
        self.merge_data = {}
        self.merge_global_data = {}
        self.metadata = {}
        self.merge_metadata = {}
        self.merge_headers = {}
        self.to_emails = []

        super().__init__(message, defaults, backend, auth=auth, *args, **kwargs)

    def get_api_endpoint(self):
        if self.sender_domain is None:
            raise AnymailError(
                "Cannot call Mailgun unknown sender domain. "
                "Either provide valid `from_email`, "
                "or set `message.esp_extra={'sender_domain': 'example.com'}`",
                backend=self.backend,
                email_message=self.message,
                payload=self,
            )
        if "/" in self.sender_domain or "%2f" in self.sender_domain.lower():
            # Mailgun returns a cryptic 200-OK "Mailgun Magnificent API" response
            # if '/' (or even %-encoded '/') confuses it about the API endpoint.
            raise AnymailError(
                "Invalid '/' in sender domain '%s'" % self.sender_domain,
                backend=self.backend,
                email_message=self.message,
                payload=self,
            )
        return "%s/messages" % quote(self.sender_domain, safe="")

    def serialize_data(self):
        self.populate_recipient_variables()
        return self.data

    # A not-so-brief digression about Mailgun's batch sending, template personalization,
    # and metadata tracking capabilities...
    #
    # Mailgun has two kinds of templates:
    #   * ESP-stored templates (handlebars syntax), referenced by template name in the
    #     send API, with substitution data supplied as "custom data" variables.
    #     Anymail's `template_id` maps to this feature.
    #   * On-the-fly templating (`%recipient.KEY%` syntax), with template variables
    #     appearing directly in the message headers and/or body, and data supplied
    #     as "recipient variables" per-recipient personalizations. Mailgun docs also
    #     sometimes refer to this data as "template variables," but it's distinct from
    #     the substitution data used for stored handelbars templates.
    #
    # Mailgun has two mechanisms for supplying additional data with a message:
    #   * "Custom data" is supplied via `v:KEY` and/or `h:X-Mailgun-Variables` fields.
    #     Custom data is passed to tracking webhooks (as 'user-variables') and is
    #     available for `{{substitutions}}` in ESP-stored handlebars templates.
    #     Normally, the same custom data is applied to every recipient of a message.
    #   * "Recipient variables" are supplied via the `recipient-variables` field, and
    #     provide per-recipient data for batch sending. The recipient specific values
    #     are available as `%recipient.KEY%` virtually anywhere in the message
    #     (including header fields and other parameters).
    #
    # Anymail needs both mechanisms to map its normalized metadata and template
    # merge_data features to Mailgun:
    # (1) Anymail's `metadata` maps directly to Mailgun's custom data, where it can be
    #     accessed from webhooks.
    # (2) Anymail's `merge_metadata` (per-recipient metadata for batch sends) maps
    #     *indirectly* through recipient-variables to Mailgun's custom data. To avoid
    #     conflicts, the recipient-variables mapping prepends 'v:' to merge_metadata
    #     keys. (E.g., Mailgun's custom-data "user" is set to "%recipient.v:user",
    #     which picks up its per-recipient value from Mailgun's
    #     `recipient-variables[to_email]["v:user"]`.)
    # (3) Anymail's `merge_data` (per-recipient template substitutions) maps directly
    #     to Mailgun's `recipient-variables`, where it can be referenced in on-the-fly
    #     templates.
    # (4) Anymail's `merge_global_data` (global template substitutions) is copied
    #     to Mailgun's `recipient-variables` for every recipient, as the default
    #     for missing `merge_data` keys.
    # (5) Only if a stored template is used, `merge_data` and `merge_global_data` are
    #     *also* mapped *indirectly* through recipient-variables to Mailgun's custom
    #     data, where they can be referenced in handlebars {{substitutions}}.
    #     (E.g., Mailgun's custom-data "name" is set to "%recipient.name%", which picks
    #     up its per-recipient value from Mailgun's
    #     `recipient-variables[to_email]["name"]`.)
    # (6) Anymail's `merge_headers` (per-recipient headers) maps to recipient-variables
    #     prepended with 'h:'.
    #
    # If Anymail's `merge_data`, `template_id` (stored templates) and `metadata` (or
    # `merge_metadata`) are used together, there's a possibility of conflicting keys
    # in Mailgun's custom data. Anymail treats that conflict as an unsupported feature
    # error.

    def populate_recipient_variables(self):
        """
        Populate Mailgun recipient-variables and custom data
        from merge data and metadata
        """
        # (numbers refer to detailed explanation above)
        # Mailgun parameters to construct:
        recipient_variables = {}
        custom_data = {}

        # (1) metadata --> Mailgun custom_data
        custom_data.update(self.metadata)

        # (2) merge_metadata --> Mailgun custom_data via recipient_variables
        if self.merge_metadata:

            def vkey(key):  # 'v:key'
                return "v:{}".format(key)

            # all keys used in any recipient's merge_metadata:
            merge_metadata_keys = flatset(
                recipient_data.keys() for recipient_data in self.merge_metadata.values()
            )
            # custom_data['key'] = '%recipient.v:key%' indirection:
            custom_data.update(
                {key: "%recipient.{}%".format(vkey(key)) for key in merge_metadata_keys}
            )
            # defaults for each recipient must cover all keys:
            base_recipient_data = {
                vkey(key): self.metadata.get(key, "") for key in merge_metadata_keys
            }
            for email in self.to_emails:
                this_recipient_data = base_recipient_data.copy()
                this_recipient_data.update(
                    {
                        vkey(key): value
                        for key, value in self.merge_metadata.get(email, {}).items()
                    }
                )
                recipient_variables.setdefault(email, {}).update(this_recipient_data)

        # (3) and (4) merge_data, merge_global_data --> Mailgun recipient_variables
        if self.merge_data or self.merge_global_data:
            # all keys used in any recipient's merge_data:
            merge_data_keys = flatset(
                recipient_data.keys() for recipient_data in self.merge_data.values()
            )
            merge_data_keys = merge_data_keys.union(self.merge_global_data.keys())
            # defaults for each recipient must cover all keys:
            base_recipient_data = {
                key: self.merge_global_data.get(key, "") for key in merge_data_keys
            }
            for email in self.to_emails:
                this_recipient_data = base_recipient_data.copy()
                this_recipient_data.update(self.merge_data.get(email, {}))
                recipient_variables.setdefault(email, {}).update(this_recipient_data)

            # (5) if template, also map Mailgun custom_data to per-recipient_variables
            if self.data.get("template") is not None:
                conflicts = merge_data_keys.intersection(custom_data.keys())
                if conflicts:
                    self.unsupported_feature(
                        "conflicting merge_data and metadata keys (%s)"
                        " when using template_id"
                        % ", ".join("'%s'" % key for key in conflicts)
                    )
                # custom_data['key'] = '%recipient.key%' indirection:
                custom_data.update(
                    {key: "%recipient.{}%".format(key) for key in merge_data_keys}
                )

        # (6) merge_headers --> Mailgun recipient_variables via 'h:'-prefixed keys
        if self.merge_headers:

            def hkey(field_name):  # 'h:Field-Name'
                return "h:{}".format(field_name.title())

            merge_header_fields = flatset(
                recipient_headers.keys()
                for recipient_headers in self.merge_headers.values()
            )
            merge_header_defaults = {
                # existing h:Field-Name value (from extra_headers), or empty string
                field: self.data.get(hkey(field), "")
                for field in merge_header_fields
            }
            self.data.update(
                # Set up 'h:Field-Name': '%recipient.h:Field-Name%' indirection
                {
                    hvar: f"%recipient.{hvar}%"
                    for hvar in [hkey(field) for field in merge_header_fields]
                }
            )

            for email in self.to_emails:
                # Each recipient's recipient_variables needs _all_ merge header fields
                recipient_headers = merge_header_defaults.copy()
                recipient_headers.update(self.merge_headers.get(email, {}))
                recipient_variables_for_headers = {
                    hkey(field): value for field, value in recipient_headers.items()
                }
                recipient_variables.setdefault(email, {}).update(
                    recipient_variables_for_headers
                )

        # populate Mailgun params
        self.data.update({"v:%s" % key: value for key, value in custom_data.items()})
        if recipient_variables or self.is_batch():
            self.data["recipient-variables"] = self.serialize_json(recipient_variables)

    #
    # Payload construction
    #

    def init_payload(self):
        self.data = {}  # {field: [multiple, values]}
        self.files = []  # [(field, multiple), (field, values)]
        self.headers = {}

    def set_from_email_list(self, emails):
        # Mailgun supports multiple From email addresses
        self.data["from"] = [email.address for email in emails]
        if self.sender_domain is None and len(emails) > 0:
            # try to intuit sender_domain from first from_email
            self.sender_domain = emails[0].domain or None

    def set_recipients(self, recipient_type, emails):
        assert recipient_type in ["to", "cc", "bcc"]
        if emails:
            self.data[recipient_type] = [email.address for email in emails]
            # used for backend.parse_recipient_status:
            self.all_recipients += emails
        if recipient_type == "to":
            # used for populate_recipient_variables:
            self.to_emails = [email.addr_spec for email in emails]

    def set_subject(self, subject):
        self.data["subject"] = subject

    def set_reply_to(self, emails):
        if emails:
            reply_to = ", ".join([str(email) for email in emails])
            self.data["h:Reply-To"] = reply_to

    def set_extra_headers(self, headers):
        for field, value in headers.items():
            self.data["h:%s" % field.title()] = value

    def set_text_body(self, body):
        self.data["text"] = body

    def set_html_body(self, body):
        if "html" in self.data:
            # second html body could show up through multiple alternatives,
            # or html body + alternative
            self.unsupported_feature("multiple html parts")
        self.data["html"] = body

    def add_alternative(self, content, mimetype):
        if mimetype.lower() == "text/x-amp-html":
            if "amp-html" in self.data:
                self.unsupported_feature("multiple html parts")
            self.data["amp-html"] = content
        else:
            super().add_alternative(content, mimetype)

    def add_attachment(self, attachment):
        # http://docs.python-requests.org/en/v2.4.3/user/advanced/#post-multiple-multipart-encoded-files
        if attachment.inline:
            field = "inline"
            name = attachment.cid
            if not name:
                self.unsupported_feature("inline attachments without Content-ID")
        else:
            field = "attachment"
            name = attachment.name
            if not name:
                self.unsupported_feature("attachments without filenames")
        self.files.append((field, (name, attachment.content, attachment.mimetype)))

    def set_envelope_sender(self, email):
        # Only the domain is used
        self.sender_domain = email.domain

    def set_metadata(self, metadata):
        self.metadata = metadata  # save for handling merge_metadata later
        for key, value in metadata.items():
            self.data["v:%s" % key] = value

    def set_send_at(self, send_at):
        # Mailgun expects RFC-2822 format dates
        # (BasePayload has converted most date-like values to datetime by now;
        # if the caller passes a string, they'll need to format it themselves.)
        if isinstance(send_at, datetime):
            send_at = rfc2822date(send_at)
        self.data["o:deliverytime"] = send_at

    def set_tags(self, tags):
        self.data["o:tag"] = tags

    def set_track_clicks(self, track_clicks):
        # Mailgun also supports an "htmlonly" option, which Anymail doesn't offer
        self.data["o:tracking-clicks"] = "yes" if track_clicks else "no"

    def set_track_opens(self, track_opens):
        self.data["o:tracking-opens"] = "yes" if track_opens else "no"

    def set_template_id(self, template_id):
        self.data["template"] = template_id

    def set_merge_data(self, merge_data):
        # Processed at serialization time (to allow merging global data)
        self.merge_data = merge_data

    def set_merge_global_data(self, merge_global_data):
        # Processed at serialization time (to allow merging global data)
        self.merge_global_data = merge_global_data

    def set_merge_metadata(self, merge_metadata):
        # Processed at serialization time (to allow combining with merge_data)
        self.merge_metadata = merge_metadata

    def set_merge_headers(self, merge_headers):
        self.merge_headers = merge_headers

    def set_esp_extra(self, extra):
        self.data.update(extra)
        # Allow override of sender_domain via esp_extra
        # (but pop it out of params to send to Mailgun)
        self.sender_domain = self.data.pop("sender_domain", self.sender_domain)


def isascii(s):
    """Returns True if str s is entirely ASCII characters.

    (Compare to Python 3.7 `str.isascii()`.)
    """
    try:
        s.encode("ascii")
    except UnicodeEncodeError:
        return False
    return True


def flatset(iterables):
    """Return a set of the items in a single-level flattening of iterables

    >>> flatset([1, 2], [2, 3])
    set(1, 2, 3)
    """
    return set(item for iterable in iterables for item in iterable)



import json
from base64 import b64encode
from unittest.mock import ANY

from django.test import tag

from anymail.inbound import AnymailInboundMessage
from anymail.signals import AnymailInboundEvent
from anymail.webhooks.mailjet import MailjetInboundWebhookView

from .utils import sample_email_content, sample_image_content
from .webhook_cases import WebhookTestCase


@tag("mailjet")
class MailjetInboundTestCase(WebhookTestCase):
    def test_inbound_basics(self):
        raw_event = {
            "Sender": "envelope-from@example.org",
            "Recipient": "test@inbound.example.com",
            # this is just the Date header from the sender, parsed to UTC:
            "Date": "20171012T013104",
            "From": '"Displayed From" <from+test@example.org>',
            "Subject": "Test subject",
            "Headers": {
                "Return-Path": [
                    "<bounce-handler=from+test%example.org@mail.example.org>"
                ],
                "Received": [
                    "from mail.example.org by parse.mailjet.com ..."
                    "by mail.example.org for <test@inbound.example.com> ...",
                    "by 10.10.1.71 with HTTP; Wed, 11 Oct 2017 18:31:04 -0700 (PDT)",
                ],
                "MIME-Version": ["1.0"],
                "From": '"Displayed From" <from+test@example.org>',
                "Date": "Wed, 11 Oct 2017 18:31:04 -0700",
                "Message-ID": "<CAEPk3R+4Zr@mail.example.org>",
                "Subject": "Test subject",
                "To": "Test Inbound <test@inbound.example.com>, other@example.com",
                "Cc": "cc@example.com",
                "Reply-To": "from+test@milter.example.org",
                "Content-Type": ['multipart/alternative; boundary="boundary0"'],
            },
            "Parts": [
                {
                    "Headers": {"Content-Type": ['text/plain; charset="UTF-8"']},
                    "ContentRef": "Text-part",
                },
                {
                    "Headers": {
                        "Content-Type": ['text/html; charset="UTF-8"'],
                        "Content-Transfer-Encoding": ["quoted-printable"],
                    },
                    "ContentRef": "Html-part",
                },
            ],
            "Text-part": "Test body plain",
            "Html-part": "<div>Test body html</div>",
            "SpamAssassinScore": "1.7",
        }

        response = self.client.post(
            "/anymail/mailjet/inbound/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=MailjetInboundWebhookView,
            event=ANY,
            esp_name="Mailjet",
        )
        # AnymailInboundEvent
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailInboundEvent)
        self.assertEqual(event.event_type, "inbound")
        # Mailjet doesn't provide inbound event timestamp:
        self.assertIsNone(event.timestamp)
        # Mailjet doesn't provide inbound event id:
        self.assertIsNone(event.event_id)
        self.assertIsInstance(event.message, AnymailInboundMessage)
        self.assertEqual(event.esp_event, raw_event)

        # AnymailInboundMessage - convenience properties
        message = event.message

        self.assertEqual(message.from_email.display_name, "Displayed From")
        self.assertEqual(message.from_email.addr_spec, "from+test@example.org")
        self.assertEqual(
            [str(e) for e in message.to],
            ["Test Inbound <test@inbound.example.com>", "other@example.com"],
        )
        self.assertEqual([str(e) for e in message.cc], ["cc@example.com"])
        self.assertEqual(message.subject, "Test subject")
        self.assertEqual(message.date.isoformat(" "), "2017-10-11 18:31:04-07:00")
        self.assertEqual(message.text, "Test body plain")
        self.assertEqual(message.html, "<div>Test body html</div>")

        self.assertEqual(message.envelope_sender, "envelope-from@example.org")
        self.assertEqual(message.envelope_recipient, "test@inbound.example.com")
        # Mailjet doesn't provide stripped plaintext body:
        self.assertIsNone(message.stripped_text)
        # Mailjet doesn't provide stripped html:
        self.assertIsNone(message.stripped_html)
        # Mailjet doesn't provide spam boolean:
        self.assertIsNone(message.spam_detected)
        self.assertEqual(message.spam_score, 1.7)

        # AnymailInboundMessage - other headers
        self.assertEqual(message["Message-ID"], "<CAEPk3R+4Zr@mail.example.org>")
        self.assertEqual(message["Reply-To"], "from+test@milter.example.org")
        self.assertEqual(
            message.get_all("Received"),
            [
                "from mail.example.org by parse.mailjet.com ..."
                "by mail.example.org for <test@inbound.example.com> ...",
                "by 10.10.1.71 with HTTP; Wed, 11 Oct 2017 18:31:04 -0700 (PDT)",
            ],
        )

    def test_attachments(self):
        image_content = sample_image_content()
        email_content = sample_email_content()
        raw_event = {
            "Headers": {
                "MIME-Version": ["1.0"],
                "Content-Type": ['multipart/mixed; boundary="boundary0"'],
            },
            "Parts": [
                {
                    "Headers": {
                        "Content-Type": ['multipart/related; boundary="boundary1"']
                    }
                },
                {
                    "Headers": {
                        "Content-Type": ['multipart/alternative; boundary="boundary2"']
                    }
                },
                {
                    "Headers": {"Content-Type": ['text/plain; charset="UTF-8"']},
                    "ContentRef": "Text-part",
                },
                {
                    "Headers": {
                        "Content-Type": ['text/html; charset="UTF-8"'],
                        "Content-Transfer-Encoding": ["quoted-printable"],
                    },
                    "ContentRef": "Html-part",
                },
                {
                    "Headers": {
                        "Content-Type": ["text/plain"],
                        "Content-Disposition": ['attachment; filename="test.txt"'],
                        "Content-Transfer-Encoding": ["quoted-printable"],
                    },
                    "ContentRef": "Attachment1",
                },
                {
                    "Headers": {
                        "Content-Type": ['image/png; name="image.png"'],
                        "Content-Disposition": ['inline; filename="image.png"'],
                        "Content-Transfer-Encoding": ["base64"],
                        "Content-ID": ["<abc123>"],
                    },
                    "ContentRef": "InlineAttachment1",
                },
                {
                    "Headers": {
                        "Content-Type": ['message/rfc822; charset="US-ASCII"'],
                        "Content-Disposition": ["attachment"],
                    },
                    "ContentRef": "Attachment2",
                },
            ],
            "Text-part": "Test body plain",
            "Html-part": "<div>Test body html <img src='cid:abc123'></div>",
            "InlineAttachment1": b64encode(image_content).decode("ascii"),
            "Attachment1": b64encode("test attachment".encode("utf-8")).decode("ascii"),
            "Attachment2": b64encode(email_content).decode("ascii"),
        }

        response = self.client.post(
            "/anymail/mailjet/inbound/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=MailjetInboundWebhookView,
            event=ANY,
            esp_name="Mailjet",
        )
        event = kwargs["event"]
        message = event.message
        attachments = message.attachments  # AnymailInboundMessage convenience accessor
        self.assertEqual(len(attachments), 2)
        self.assertEqual(attachments[0].get_filename(), "test.txt")
        self.assertEqual(attachments[0].get_content_type(), "text/plain")
        self.assertEqual(attachments[0].get_content_text(), "test attachment")
        self.assertEqual(attachments[1].get_content_type(), "message/rfc822")
        self.assertEqualIgnoringHeaderFolding(
            attachments[1].get_content_bytes(), email_content
        )

        inlines = message.content_id_map
        self.assertEqual(len(inlines), 1)
        inline = inlines["abc123"]
        self.assertEqual(inline.get_filename(), "image.png")
        self.assertEqual(inline.get_content_type(), "image/png")
        self.assertEqual(inline.get_content_bytes(), image_content)



from __future__ import annotations

import hashlib
import json
from copy import deepcopy
from datetime import datetime, timezone
from unittest.mock import ANY

from django.test import override_settings, tag

from anymail.exceptions import AnymailConfigurationError
from anymail.signals import AnymailTrackingEvent
from anymail.webhooks.unisender_go import UnisenderGoTrackingWebhookView

from .utils import test_file_content
from .webhook_cases import WebhookBasicAuthTestCase, WebhookTestCase

TEST_API_KEY = "TEST_API_KEY"


def unisender_go_signed_payload(
    data: dict, api_key: str, json_options: dict | None = None
) -> bytes:
    """
    Return data serialized to JSON and signed with api_key, using Unisender Go's
    webhook signature in the top-level "auth" field:

        "MD5 hash of the string body of the message, with the auth value replaced
        by the api_key of the user/project whose handler is being called."
        https://godocs.unisender.ru/web-api-ref#callback-format

    Any json_options are passed to json.dumps as kwargs.

    This modifies data to add the "auth" field.
    """
    json_options = json_options or {"separators": (",", ":")}
    placeholder = "__PLACEHOLDER_FOR_SIGNATURE__"
    data["auth"] = placeholder
    serialized_data = json.dumps(data, **json_options)
    signature = hashlib.md5(
        serialized_data.replace(placeholder, api_key).encode()
    ).hexdigest()
    signed_data = serialized_data.replace(placeholder, signature)
    data["auth"] = signature  # make available to the caller
    return signed_data.encode()


class UnisenderGoWebhookTestCase(WebhookTestCase):
    def client_post_signed(
        self,
        url: str,
        data: dict,
        api_key: str = TEST_API_KEY,
        json_options: dict | None = None,
        **kwargs,
    ):
        """
        Return self.client.post(url, serialized json_data) signed with api_key
        using json_options.

        Additional kwargs are passed to self.client.post()
        """
        signed_data = unisender_go_signed_payload(data, api_key, json_options)
        return self.client.post(
            url, content_type="application/json", data=signed_data, **kwargs
        )


@tag("unisender_go")
class UnisenderGoWebhookSettingsTestCase(UnisenderGoWebhookTestCase):
    def test_requires_api_key(self):
        with self.assertRaisesMessage(
            AnymailConfigurationError, "UNISENDER_GO_API_KEY"
        ):
            self.client_post_signed("/anymail/unisender_go/tracking/", {})

    @override_settings(ANYMAIL={"UNISENDER_GO_API_KEY": "SETTINGS_API_KEY"})
    def test_view_params_api_key_override(self):
        """Webhook api_key can be provided as a view param"""
        view = UnisenderGoTrackingWebhookView.as_view(api_key="VIEW_API_KEY")
        view_instance = view.view_class(**view.view_initkwargs)
        self.assertEqual(view_instance.api_key_bytes, b"VIEW_API_KEY")


@tag("unisender_go")
@override_settings(
    # (Use expanded setting name because WebhookBasicAuthTestCase sets ANYMAIL={}.)
    ANYMAIL_UNISENDER_GO_API_KEY=TEST_API_KEY,
)
class UnisenderGoWebhookSecurityTestCase(
    UnisenderGoWebhookTestCase, WebhookBasicAuthTestCase
):
    should_warn_if_no_auth = False  # because we check webhook signature

    payload = {
        # "auth" is added by self.client_post_signed()
        "events_by_user": [
            {
                "user_id": 123456,
                "events": [
                    {
                        "event_name": "transactional_email_status",
                        "event_data": {
                            "job_id": "1sn15Z-0007Le-GtVN",
                            "email": "test@example.com",
                            "status": "sent",
                            "metadata": {"can_be_unicode": "Метаданные"},
                            "event_time": "2024-09-07 19:27:17",
                        },
                    }
                ],
            }
        ]
    }

    def call_webhook(self):
        return self.client_post_signed("/anymail/unisender_go/tracking/", self.payload)

    # Additional tests are in WebhookBasicAuthTestCase

    def test_verifies_correct_signature(self):
        response = self.client_post_signed(
            "/anymail/unisender_go/tracking/", self.payload
        )
        self.assertEqual(response.status_code, 200)

    def test_rejects_bad_signature(self):
        # This also verifies that the error log references the correct setting to check.
        with self.assertLogs() as logs:
            response = self.client_post_signed(
                "/anymail/unisender_go/tracking/",
                self.payload,
                api_key="OTHER_API_KEY",
            )
        # SuspiciousOperation causes 400 response (even in test client):
        self.assertEqual(response.status_code, 400)
        self.assertIn("check Anymail UNISENDER_GO_API_KEY", logs.output[0])
        self.assertNotIn("Project ID", logs.output[0])

    def test_rejects_missing_signature(self):
        payload = deepcopy(self.payload)
        del payload["auth"]
        # Post directly (without signing to add auth):
        response = self.client.post(
            "/anymail/unisender_go/tracking/",
            content_type="application/json",
            data=json.dumps(payload),
        )
        self.assertEqual(response.status_code, 400)

    def test_rejects_problem_signatures(self):
        # Make sure our `body.replace(auth, key)` approach can't be confused
        # by invalid payloads.
        for bad_auth in ["", " ", ":", "{", '"', 0, None, [], {}]:
            with self.subTest(auth=bad_auth):
                payload = deepcopy(self.payload)
                payload["auth"] = bad_auth
                response = self.client.post(
                    "/anymail/unisender_go/tracking/",
                    content_type="application/json",
                    data=json.dumps(payload),
                )
                self.assertEqual(response.status_code, 400)

    def test_error_includes_project_id(self):
        # If the webhook has a selected project, mention
        # its id in the validation error to assist in debugging.
        payload = deepcopy(self.payload)
        payload["events_by_user"][0].update(
            {"project_id": 999999, "project_name": "Test project"}
        )
        with self.assertLogs() as logs:
            response = self.client_post_signed(
                "/anymail/unisender_go/tracking/",
                payload,
                api_key="OTHER_API_KEY",
            )
        self.assertEqual(response.status_code, 400)
        self.assertIn(
            "check Anymail UNISENDER_GO_API_KEY setting is for Project ID 999999",
            logs.output[0],
        )

    def test_error_includes_project_id_single_event(self):
        # Selected project works with "single event" option.
        payload = {
            "user_id": 123456,
            "project_id": 999999,
            "project_name": "Test project",
            "event_name": "transactional_email_status",
            "job_id": "1sn15Z-0007Le-GtVN",
            "email": "test@example.com",
            "status": "sent",
            "event_time": "2024-09-07 19:27:17",
        }
        with self.assertLogs() as logs:
            response = self.client_post_signed(
                "/anymail/unisender_go/tracking/",
                payload,
                api_key="OTHER_API_KEY",
            )
        self.assertEqual(response.status_code, 400)
        self.assertIn(
            "check Anymail UNISENDER_GO_API_KEY setting is for Project ID 999999",
            logs.output[0],
        )

    def test_insensitive_to_json_serialization_options(self):
        # Our webhook signature verification must not depend on the exact
        # details of how Unisender Go serializes the JSON payload.
        for json_options in [
            {"separators": None},
            {"ensure_ascii": False},
            {"indent": 4},
            {"sort_keys": True},
        ]:
            with self.subTest(options=json_options):
                response = self.client_post_signed(
                    "/anymail/unisender_go/tracking/",
                    self.payload,
                    json_options=json_options,
                )
                self.assertEqual(response.status_code, 200)

    @override_settings(
        # noqa: secret-scanning: this API key has been disabled
        ANYMAIL={"UNISENDER_GO_API_KEY": "6mjstx9gwi7qj8eni6m77hfiiw6aifmss154y4ze"}
    )
    def test_actual_signed_payload(self):
        # Test our signature verification using an actual payload and API key.
        payload = test_file_content("unisender-go-tracking-test-payload.json.raw")
        # (If an editor or pre-commit forces a trailing newline, the test breaks.)
        assert payload[-1] != b"\n", "Test payload must not have end-of-file newline"
        response = self.client.post(
            "/anymail/unisender_go/tracking/",
            content_type="application/json",
            data=payload,
        )
        self.assertEqual(response.status_code, 200)


@tag("unisender_go")
@override_settings(ANYMAIL={"UNISENDER_GO_API_KEY": TEST_API_KEY})
class UnisenderGoTestCase(UnisenderGoWebhookTestCase):
    # Most of these tests use Unisender Go's "single event" option for brevity.
    # Anymail also supports (and recommends) multiple event webhook option;
    # tests for that are toward the end.

    def test_sent_event(self):
        raw_event = {
            "event_name": "transactional_email_status",
            "user_id": 111111,
            "project_id": 999999,
            "project_name": "Testing",
            "job_id": "1smi9f-00057m-86zr",
            "metadata": {
                "anymail_id": "00001111-2222-3333-4444-555566667777",
                "cohort": "group a121",
            },
            "email": "recipient@example.com",
            "status": "sent",
            "event_time": "2024-09-06 23:14:19",
        }
        response = self.client_post_signed("/anymail/unisender_go/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=UnisenderGoTrackingWebhookView,
            event=ANY,
            esp_name="Unisender Go",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "sent")
        self.assertEqual(
            event.timestamp,
            datetime(2024, 9, 6, 23, 14, 19, tzinfo=timezone.utc),
        )
        # event.message_id matches the message.anymail_status.message_id
        # from when the message was sent. It comes from metadata.anymail_id
        # if present (added by UNISENDER_GO_GENERATE_MESSAGE_ID).
        self.assertEqual(event.message_id, "00001111-2222-3333-4444-555566667777")
        # Unisender Go does not include a useful event_id
        self.assertIsNone(event.event_id)
        self.assertEqual(event.recipient, "recipient@example.com")
        # Although Unisender Go's email-send docs claim tags are sent to webhooks,
        # its webhook docs don't show tags (and they aren't actually sent 9/2024).
        #   self.assertEqual(event.tags, ["tag1", "Tag 2"])
        # Our added "anymail_id" should be removed from metadata.
        self.assertEqual(event.metadata, {"cohort": "group a121"})
        self.assertEqual(event.esp_event, raw_event)

    def test_delivered_event(self):
        raw_event = {
            "event_name": "transactional_email_status",
            "user_id": 111111,
            "job_id": "1smi9f-00057m-86zr",
            "email": "recipient@example.com",
            "status": "delivered",
            "event_time": "2024-09-06 23:14:24",
        }
        response = self.client_post_signed("/anymail/unisender_go/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=UnisenderGoTrackingWebhookView,
            event=ANY,
            esp_name="Unisender Go",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "delivered")
        # If UNISENDER_GO_GENERATE_MESSAGE_ID not enabled, Unisender Go's
        # job_id becomes Anymail's message_id.
        self.assertEqual(event.message_id, "1smi9f-00057m-86zr")
        self.assertEqual(event.recipient, "recipient@example.com")

    def test_hard_bounced_event(self):
        raw_event = {
            "event_name": "transactional_email_status",
            "status": "hard_bounced",
            "email": "bounce@example.com",
            "delivery_info": {
                "delivery_status": "err_user_unknown",
                "destination_response": "555 5.7.1 User unknown 'bounce@example.com'.",
            },
            "event_time": "2024-09-06 23:22:40",
        }
        response = self.client_post_signed("/anymail/unisender_go/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=UnisenderGoTrackingWebhookView,
            event=ANY,
            esp_name="Unisender Go",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.recipient, "bounce@example.com")
        self.assertEqual(event.reject_reason, "bounced")
        self.assertEqual(event.description, "err_user_unknown")
        self.assertEqual(
            event.mta_response, "555 5.7.1 User unknown 'bounce@example.com'."
        )

    def test_soft_bounced_event(self):
        raw_event = {
            "event_name": "transactional_email_status",
            "status": "soft_bounced",
            "email": "full@example.com",
            "delivery_info": {
                "delivery_status": "err_mailbox_full",
                "destination_response": "554 5.2.2 Mailbox full 'full@example.com'.",
            },
            "event_time": "2024-09-06 23:22:40",
        }
        response = self.client_post_signed("/anymail/unisender_go/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=UnisenderGoTrackingWebhookView,
            event=ANY,
            esp_name="Unisender Go",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "deferred")
        self.assertEqual(event.recipient, "full@example.com")
        self.assertEqual(event.reject_reason, "bounced")
        self.assertEqual(event.description, "err_mailbox_full")
        self.assertEqual(
            event.mta_response, "554 5.2.2 Mailbox full 'full@example.com'."
        )

    def test_spam_event(self):
        raw_event = {
            "event_name": "transactional_email_status",
            "status": "spam",
            "email": "to@example.com",
            "delivery_info": {
                "delivery_status": "err_spam_rejected",
                "destination_response": "550 Spam rejected",
            },
        }
        response = self.client_post_signed("/anymail/unisender_go/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=UnisenderGoTrackingWebhookView,
            event=ANY,
            esp_name="Unisender Go",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "complained")
        self.assertEqual(event.recipient, "to@example.com")
        self.assertEqual(event.description, "err_spam_rejected")
        self.assertEqual(event.mta_response, "550 Spam rejected")

    def test_unsubscribed_event(self):
        raw_event = {
            "event_name": "transactional_email_status",
            "status": "unsubscribed",
            "email": "to@example.com",
            "comment": "From unsubscribe page 'comment' field",
        }
        response = self.client_post_signed("/anymail/unisender_go/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=UnisenderGoTrackingWebhookView,
            event=ANY,
            esp_name="Unisender Go",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "unsubscribed")
        self.assertEqual(event.recipient, "to@example.com")
        self.assertEqual(event.description, "From unsubscribe page 'comment' field")

    def test_opened_event(self):
        raw_event = {
            "event_name": "transactional_email_status",
            "status": "opened",
            "email": "to@example.com",
            "delivery_info": {
                "user_agent": "... via ggpht.com GoogleImageProxy",
                "ip": "10.10.1.333",
            },
        }
        response = self.client_post_signed("/anymail/unisender_go/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=UnisenderGoTrackingWebhookView,
            event=ANY,
            esp_name="Unisender Go",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "opened")
        self.assertEqual(event.recipient, "to@example.com")
        self.assertEqual(event.user_agent, "... via ggpht.com GoogleImageProxy")

    def test_clicked_event(self):
        raw_event = {
            "event_name": "transactional_email_status",
            "status": "clicked",
            "email": "to@example.com",
            "url": "https://example.com",
            "delivery_info": {
                "user_agent": "Mozilla/5.0 AppleWebKit/537.36 ...",
                "ip": "192.168.1.333",
            },
        }
        response = self.client_post_signed("/anymail/unisender_go/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=UnisenderGoTrackingWebhookView,
            event=ANY,
            esp_name="Unisender Go",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "clicked")
        self.assertEqual(event.recipient, "to@example.com")
        self.assertEqual(event.click_url, "https://example.com")
        self.assertEqual(event.user_agent, "Mozilla/5.0 AppleWebKit/537.36 ...")

    def test_multiple_event_option(self):
        # Payload format is different when "Use single event" not checked.
        raw_event = {
            "events_by_user": [
                {
                    "user_id": 111111,
                    "events": [
                        {
                            "event_name": "transactional_email_status",
                            "event_data": {
                                "job_id": "1sn15Z-0007Le-GtVN",
                                "email": "to@example.com",
                                "status": "sent",
                                "event_time": "2024-09-07 19:27:17",
                            },
                        },
                        {
                            "event_name": "transactional_email_status",
                            "event_data": {
                                "job_id": "1sn15Z-0007Le-GtVN",
                                "email": "cc@example.com",
                                "status": "delivered",
                                "event_time": "2024-09-07 19:27:17",
                            },
                        },
                    ],
                }
            ],
        }
        response = self.client_post_signed("/anymail/unisender_go/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(self.tracking_handler.call_count, 2)
        events = [
            kwargs["event"] for (args, kwargs) in self.tracking_handler.call_args_list
        ]
        self.assertEqual(events[0].event_type, "sent")
        self.assertEqual(events[0].recipient, "to@example.com")
        self.assertEqual(events[1].event_type, "delivered")
        self.assertEqual(events[1].recipient, "cc@example.com")

        # esp_event is event_data for each event
        self.assertEqual(
            events[0].esp_event,
            raw_event["events_by_user"][0]["events"][0]["event_data"],
        )
        self.assertEqual(
            events[1].esp_event,
            raw_event["events_by_user"][0]["events"][1]["event_data"],
        )

    def test_webhook_setup_verification(self):
        # Unisender Go verifies webhook at setup time by calling GET.
        response = self.client.get("/anymail/unisender_go/tracking/")
        self.assertEqual(response.status_code, 200)
